#![no_std]

/// # OFW_TRUE
/// a constant representing the OpenFirmware "true" value
pub const OFW_TRUE: i32 = -1;

/// # OFW_FALSE
/// a constant representing the OpenFirmware "false" value
pub const OFW_FALSE: i32 = 0;

/// # OFWBoolean
/// a type representing the OpenFirmware boolean type
/// can be either OFW_TRUE or OFW_FALSE
pub type OFWBoolean = i32;

/// # PHandle
/// handle to a "package" in OpenFirmware
/// usually used to access properties
pub type PHandle = i32;

/// # IHandle
/// handle to a package instance in OpenFirmware
/// usually used to call methods from a package
pub type IHandle = i32;

use crate::arch::EntryFunction;
use crate::ntstr::NTSTR;

/// # ntstr
/// module for handling null-terminated strings
pub mod ntstr;

/// # arch
/// module for architecture-specific code
pub mod arch;

/// # cell_conv
/// module for converting between OpenFirmware cells and Rust integers
pub mod cell_conv;

#[cfg(test)]
mod tests;

/// # Args
/// a base structure for all calls to the OpenFirmware entry function,
/// shouldn't be used directly but rather used as a base for other "service" structures
/// an example "service struct" is as follows
/// ```
/// use ofw::{Args, OFWBoolean};
/// use ofw::ntstr::NTSTR;
/// /// # test
/// /// nargs: 1
/// /// nrets: 1
/// /// takes a service name string, and returns OFW_TRUE if the service is missing or OFW_FALSE if it is present
/// #[repr(C)]
/// pub struct TestService {
///    pub args: Args,
///    pub service: NTSTR,
///    pub missing: OFWBoolean,
/// }
/// ```
#[repr(C)]
pub struct Args {
    pub service: NTSTR,
    // ieee 1275, 2.4.1 General conventions: "All numbers on the stack are signed integers, _32 bits[...]"
    pub nargs: i32,
    pub nrets: i32,
    // inheriting structs will continue from here
}

/// # call!
/// a macro for calling OpenFirmware services
/// ## example
/// ```
/// use ofw::{call, service_result, ServiceResult};
/// use ofw::arch::EntryFunction;
/// let entry_fn: EntryFunction = unimplemented!("get the entry function from somewhere");
/// // <entry function>, <service name>, <# of args>, <# of rets>, <args...>
/// let (success, returns): (bool, [i32; 1]) = call!(entry_fn, ntstr!("test"), 1, 1, ntstr!("test"));
/// if success {
///    let test_result: i32 = returns[0];
/// }
/// ```
#[macro_export]
macro_rules! call {
    ($entry_fn:expr, $service:expr, $nargs:expr, $nrets:expr, $($args:expr),*) => {
        {
            let mut args_and_rets: [i32; $nargs as usize + $nrets as usize] = [0; $nargs as usize + $nrets as usize];
            let mut rets: [i32; $nrets as usize] = [0; $nrets as usize];
            let mut i = 0;
            #[allow(unused_assignments)]
            {
                $(
                    args_and_rets[i] = $args.into();
                    i += 1;
                )*
            }
            let success: (bool, [i32; $nargs as usize + $nrets as usize]) = $crate::ofw_call($entry_fn, $service, $nargs, $nrets, args_and_rets);
            for i in 0..$nrets as usize {
                rets[i] = success.1[i + $nargs as usize];
            }
            (success.0, rets)
        }
    };
}

/// # call_method!
/// a macro for calling package methods
/// ## example
/// ```
/// use ofw::arch::EntryFunction;
/// use ofw::{call_method, method_service_result, ServiceResult};
///
/// let entry_fn: EntryFunction = unimplemented!("get the entry function from somewhere");
/// let package_instance: i32 = unimplemented!("get the package instance from somewhere");
/// // <entry function>, <package instance>, <method name>, <# of args>, <# of rets>, <args...>
/// let (success, returns): (bool, [i32; 1]) = call_method!(entry_fn, package_instance, ntstr!("seek"), 2, 1, 0, 0);
/// if success {
///   let seek_result: i32 = returns[0];
/// }
/// ```
#[macro_export]
macro_rules! call_method {
    ($entry_fn:expr, $package_instance:expr, $service:expr, $nargs:expr, $nrets:expr, $($args:expr),*) => {
        {
            let mut args_and_rets: [i32; $nargs as usize + $nrets as usize + 1] = [0; $nargs as usize + $nrets as usize + 1];
            let mut rets: [i32; $nrets as usize] = [0; $nrets as usize];
            let mut i = 0;
            #[allow(unused_assignments)]
            {
                $(
                    args_and_rets[i] = $args.into();
                    i += 1;
                )*
            }
            let success: (bool, [i32; $nargs as usize + $nrets as usize + 1]) = $crate::ofw_call_method($entry_fn, $package_instance, $service, $nargs, $nrets, args_and_rets);
            for i in 0..$nrets as usize {
                rets[i] = success.1[i + $nargs as usize + 1];
            }
            (success.0 && success.1[$nargs as usize] != 0, rets)
        }
    };
}

/// # ofw_call
/// a function to call any OpenFirmware-provided service
/// usage of the macro `call!` is preferred, but this function can be used directly
/// ## example
/// ```
/// use ofw::arch::EntryFunction;
/// use ofw::{ntstr, ofw_call};
/// let entry_fn: EntryFunction = unimplemented!("get the entry function from somewhere");
/// let args: [i32; 2] = [ntstr!("test") as i32, 0];
/// let (was_service_successfully_called, args): (bool, [i32; 2]) = ofw_call(entry_fn, ntstr!("test"), 1, 1, args);
/// let test_result: i32 = args[1];
/// ```
pub fn ofw_call<T: Sized>(entry_fn: EntryFunction, service: NTSTR, num_args: i32, num_rets: i32, args_and_rets: T) -> (bool, T) {
    #[repr(C)]
    struct GenericArgs<U: Sized> {
        args: Args,
        buffer: U,
    }
    let mut generic_args = GenericArgs {
        args: Args {
            service,
            nrets: num_rets,
            nargs: num_args,
        },
        buffer: args_and_rets,
    };

    let result: bool;

    #[cfg(target_arch = "powerpc")]
    {
        result = entry_fn(&mut generic_args as *mut _ as *mut Args) != -1;
    }
    #[cfg(target_arch = "x86_64")]
    {
        result = entry_fn(&mut generic_args as *mut _ as *mut Args) != -1;
    }

    (result, generic_args.buffer)
}

/// # ofw_call_method
/// a function to call a method provided by a package instance
/// usage of the `call_method!` macro is preferred, but this function can be used directly
/// ## example
/// ```
/// // we will be calling the "seek" method on an imaginary block device
/// // the seek method takes 2 arguments, and returns 1 value
/// // addr_hi, addr_lo -- okay?
/// // however, we must add another cell in order to account for the catch-result
/// // (see ieee 1275, 6.3.2.2 Device tree)
/// use ofw::{IHandle, ntstr, ofw_call_method};
/// use ofw::arch::EntryFunction;
/// let entry_fn: EntryFunction = unimplemented!("get the entry function from somewhere");
/// let instance: IHandle = unimplemented!("get the package instance from somewhere");
/// // addr_hi, addr_lo, catch-result, okay?
/// let args: [i32; 4] = [0, 0, 0, 0];
/// // note that num_args and num_rets are not the same as the length of args, this is intentional
/// let (was_service_successfully_called, args): (bool, [i32; 4]) = ofw_call_method(entry_fn, instance, ntstr!("seek"), 2, 1, args);
/// if was_service_successfully_called && args[2] != 0 {
///     let seek_result: i32 = args[3];
/// }
/// ```
pub fn ofw_call_method<T>(entry_fn: EntryFunction, ihandle: IHandle, method: NTSTR, num_args: i32, num_rets: i32, args_and_rets: T) -> (bool, T) {
    #[repr(C)]
    struct CallMethodArgs<U> {
        args: Args,
        method: NTSTR,
        ihandle: IHandle,
        buffer: U,
    }

    let mut call_method_args = CallMethodArgs {
        args: Args {
            service: ntstr!("call-method"),
            nrets: num_rets + 1,
            nargs: num_args + 2,
        },
        method,
        ihandle,
        buffer: args_and_rets,
    };

    let result: bool;

    #[cfg(target_arch = "powerpc")]
    {
        result = entry_fn(&mut call_method_args as *mut _ as *mut Args) != OFW_TRUE;
    }
    #[cfg(target_arch = "x86_64")]
    {
        result = entry_fn(&mut call_method_args as *mut _ as *mut Args) != OFW_TRUE;
    }

    (result, call_method_args.buffer)
}