use crate::arch::EntryFunction;
use crate::{Args, call, call_method, IHandle, ntstr, OFW_FALSE, OFW_TRUE};
use crate::cell_conv::DoubleWord;
use crate::ntstr::NTSTR;

#[test]
fn ntstr_macro() {
    let a = ntstr!("Hello, world!");
    let b = NTSTR::new(b"Hello, world!\0".as_ptr());
    assert_eq!(a, b);
}

#[test]
fn doubleword_conversion() {
    let a: u64 = 0xabcdef0123456789;
    let a_dw: DoubleWord = a.into();
    let b: u64 = a_dw.into();
    assert_eq!(a, b);
    assert_eq!(a_dw.high as u32, 0xabcdef01);
    assert_eq!(a_dw.low as u32, 0x23456789);
}

#[test]
fn call_ofw() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct GenericArgs {
            args: Args,
            buffer: [i32; 2],
        }
        let args = unsafe { &mut *(args as *mut GenericArgs) };
        assert_eq!(args.args.service, ntstr!("test"));
        assert_eq!(args.args.nargs, 1);
        assert_eq!(args.args.nrets, 1);
        args.buffer[1] = OFW_FALSE; // not missing
        0
    }

    let entry_fn: EntryFunction = test_entry_function;
    let (success, results) = call!(entry_fn, ntstr!("test"), 1, 1, ntstr!("test"));
    assert!(success);
    assert_eq!(results[0], OFW_FALSE);
}


#[test]
fn call_ofw_missing() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct GenericArgs {
            args: Args,
            buffer: [i32; 2],
        }
        let args = unsafe { &mut *(args as *mut GenericArgs) };
        assert_eq!(args.args.service, ntstr!("test"));
        assert_eq!(args.args.nargs, 1);
        assert_eq!(args.args.nrets, 1);
        args.buffer[1] = OFW_TRUE; // missing
        0
    }

    let entry_fn: EntryFunction = test_entry_function;
    let (success, results) = call!(entry_fn, ntstr!("test"), 1, 1, ntstr!("aioshjfoahwfi"));
    assert!(success);
    assert_eq!(results[0], OFW_TRUE);
}

#[test]
fn call_ofw_failure() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct GenericArgs {
            args: Args
        }
        let args = unsafe { &mut *(args as *mut GenericArgs) };
        assert_eq!(args.args.service, ntstr!("this-method-does-not-exist"));
        assert_eq!(args.args.nargs, 1);
        assert_eq!(args.args.nrets, 1);
        -1
    }

    let entry_fn: EntryFunction = test_entry_function;
    let (success, _) = call!(entry_fn, ntstr!("this-method-does-not-exist"), 1, 1, ntstr!("aioshjfoahwfi"));
    assert!(!success);
}

#[test]
fn call_method_ofw() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct CallMethodArgs {
            args: Args,
            method: NTSTR,
            ihandle: IHandle,
            buffer: [i32; 4],
        }
        let args = unsafe { &mut *(args as *mut CallMethodArgs) };
        assert_eq!(args.args.service, ntstr!("call-method"));
        assert_eq!(args.method, ntstr!("seek"));
        assert_eq!(args.ihandle, 0);
        assert_eq!(args.args.nargs, 2 + 2);
        assert_eq!(args.args.nrets, 1 + 1);
        // 2 offset from args in buffer
        args.buffer[2] = 1; // set catch-result to non-zero
        args.buffer[2 + 1] = OFW_TRUE; // set okay? to true
        0
    }

    let entry_fn: EntryFunction = test_entry_function;
    let ih: IHandle = 0;
    let (success, returns) = call_method!(entry_fn, ih, ntstr!("seek"), 2, 1, 0, 0);
    assert!(success);
    assert_eq!(returns[0], OFW_TRUE);
}

#[test]
fn call_method_ofw_seek_failure() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct CallMethodArgs {
            args: Args,
            method: NTSTR,
            ihandle: IHandle,
            buffer: [i32; 4],
        }
        let args = unsafe { &mut *(args as *mut CallMethodArgs) };
        assert_eq!(args.args.service, ntstr!("call-method"));
        assert_eq!(args.method, ntstr!("seek"));
        assert_eq!(args.ihandle, 0);
        assert_eq!(args.args.nargs, 2 + 2);
        assert_eq!(args.args.nrets, 1 + 1);
        // 2 offset from args in buffer
        args.buffer[2] = 1; // set catch-result to non-zero
        args.buffer[2 + 1] = OFW_FALSE; // set okay? to false
        0
    }

    let entry_fn: EntryFunction = test_entry_function;
    let ih: IHandle = 0;
    let (success, returns) = call_method!(entry_fn, ih, ntstr!("seek"), 2, 1, 0, 0);
    assert!(success);
    assert_eq!(returns[0], OFW_FALSE);
}

#[test]
fn call_method_ofw_failure() {
    pub extern "C" fn test_entry_function(args: *mut Args) -> i32 {
        #[repr(C)]
        struct CallMethodArgs {
            args: Args,
            method: NTSTR,
            ihandle: IHandle,
            buffer: [i32; 4],
        }
        let args = unsafe { &mut *(args as *mut CallMethodArgs) };
        assert_eq!(args.args.service, ntstr!("call-method"));
        assert_eq!(args.method, ntstr!("asiofhawoptgh"));
        assert_eq!(args.ihandle, 0);
        assert_eq!(args.args.nargs, 2 + 2);
        assert_eq!(args.args.nrets, 1 + 1);
        // 2 offset from args in buffer
        args.buffer[2] = 0; // set catch-result to 0 (caught error, package doesn't have this method)
        0
    }

    let entry_fn: EntryFunction = test_entry_function;
    let ih: IHandle = 0;
    let (success, _) = call_method!(entry_fn, ih, ntstr!("asiofhawoptgh"), 2, 1, 0, 0);
    assert!(!success);
}